/*
 
  copyright (c) 2004, Michal Salaban <emes@pld-linux.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2 as
  published by the Free Software Foundation (see file COPYING for details).

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
*/

/* $Id: getstats.c,v 1.14 2004/05/29 12:55:09 emes Exp $ */

#include "sys/exit.h"
#include "parseopt/lex.h"
#include "datastruct.h"
#include "getstats.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/vfs.h>

#ifdef DEBUG
#include <syslog.h>
#include "sys/log.h"
#endif

sf_value get_stat_fs (sf_stat_def *def, sf_list *fslist)
{
    sf_value        res;
    sf_fs_stats     *fs = get_fs_entry_from_list (fslist, def->arg[1].path);
    sf_res_state    slctr = def->arg[0].resstat;

#ifdef DEBUG
    syslog (LOG_DEBUG,
            "get_stat_fs(%s): total = %ld, used = %ld, free = %ld, avail = %ld",
            def->arg[1].path,
            fs->val[VA_TOTAL],
            fs->val[VA_USED],
            fs->val[VA_FREE],
            fs->val[VA_AVAIL]
            );
#endif

    res.type = INTEGER;
    res.ptr = &(fs->val[slctr]);
    return res;
}

sf_value get_stat_proc (sf_stat_def *def, void *proctab)
{
    sf_value    res;
    
    bail_out (EXIT_BUG, "get_stat_proc(): NIY");
    return res;
}

sf_value get_stat_value (sf_stat_def *def)
{
    sf_value        res;
    sf_res_state    selector = def->arg[0].resstat;

#ifdef DEBUG
    syslog (LOG_DEBUG, "get_stat_value(%s)", def_2_string(def));
#endif

    switch (def->label) {
        case ST_PROC:
            res = get_stat_proc (def, main_db->proc);
            break;
        case ST_FS:
            res = get_stat_fs (def, main_db->fs);
            break;
        case ST_MEM:
            res.type = INTEGER;
            res.ptr = (void *) &(main_db->mem[selector]);
            break;
        case ST_SWAP:
            res.type = INTEGER;
            res.ptr = (void *) &(main_db->swap[selector]);
            break;
        case ST_LOAD:
            res.type = DOUBLE;
            switch (def->arg[0].laminutes) {
                case 1:
                    res.ptr = (void *) &(main_db->load[0]);
                    break;
                case 5:
                    res.ptr = (void *) &(main_db->load[1]);
                    break;
                case 15:
                    res.ptr = (void *) &(main_db->load[2]);
                    break;
                default:
                    bail_out (EXIT_BUG, "get_stat_value(): invalid minutes for LA");
            }
            break;
        default:
            bail_out (EXIT_BUG, "get_stat_value(): invalid stat family");
    }
    return res;
}

const char      *lafile = "/proc/loadavg";
int             lafha;
const char      *memfile = "/proc/meminfo";
int             memfha;
    
char            fbuf[ BUFSIZE ];

void fetch_la (sf_database *db)
{
    char        *one, *two, *three;

    lseek (lafha, 0, SEEK_SET);
    read (lafha, (char *) &fbuf, BUFSIZE);

    one   = &fbuf[0];
    two   = index (one, ' ') + 1;
    three = index (two, ' ') + 1;
    
    db->load[0] = atof (one);
    db->load[1] = atof (two);
    db->load[2] = atof (three);

#ifdef DEBUG
    printf ("fetch_la(): la1=%0.2f la5=%0.2f la15=%0.2f\n",
            db->load[0], db->load[1], db->load[2]);
#endif
}

void fetch_mem (sf_database *db)
{
    char                *ptr;
    long int            vmtot, vmused, vmbuf, vmcache;

    lseek (memfha, 0, SEEK_SET);
    read (memfha, (char *) &fbuf, BUFSIZE);

    /* scanning /proc/meminfo for needed values */

    /* 1. memory total is in second line after 'Mem:\ *' */
    ptr = fto_notspace (fto_space (fto_newline ((char *)&fbuf)));
    vmtot = atol (ptr);
    /* 2. used mem is next value */
    ptr = fto_notspace (fto_space (ptr));
    vmused = atol (ptr);
    /* 3. then there are: 'free' field, 'shared' field,
     *    and what we're looking for - buffers
     */
    ptr = fto_notspace (fto_space (
                fto_notspace (fto_space (
                        fto_notspace (fto_space (ptr))
                        ))
                ));
    vmbuf = atol (ptr);
    /* 4. cached is next field */
    ptr = fto_notspace (fto_space (ptr));
    vmcache = atol (ptr);

    /* 5. next line, swap values. first is 'Swap:', then total. */
    ptr = fto_notspace (fto_space (fto_newline (ptr)));
    db->swap[VA_TOTAL] = atol (ptr);

    /* 6. next one is swap-used */
    ptr = fto_notspace (fto_space (ptr));
    db->swap[VA_USED] = atol (ptr);
    
    /* 7. swap-free */
    ptr = fto_notspace (fto_space (ptr));
    db->swap[VA_FREE] = atol (ptr);
    db->swap[VA_AVAIL] = db->swap[VA_FREE];

    /* calculate real values, i.e. substract buffers and cache size */
    db->mem[VA_TOTAL] = vmtot;
    db->mem[VA_USED] = vmused - (vmbuf + vmcache);
    db->mem[VA_FREE] = vmtot - db->mem[VA_USED];
    db->mem[VA_AVAIL] = db->mem[VA_FREE];

#ifdef DEBUG
    printf ("fetch_mem(): memfree=%d memused=%d swapfree=%d swapused=%d\n",
            db->mem[VA_FREE],
            db->mem[VA_USED],
            db->swap[VA_FREE],
            db->swap[VA_USED]
            );
#endif
}

void fetch_pathspace (sf_fs_stats *fs) 
{
    struct statfs   buf;
    long int        bsizeKB;

    statfs (fs->path, &buf);

    /* block size in KB */
    bsizeKB = buf.f_bsize / 1024;
 
    fs->val[VA_TOTAL] = buf.f_blocks * bsizeKB;
    fs->val[VA_AVAIL] = buf.f_bavail * bsizeKB;
    fs->val[VA_FREE]  = buf.f_bfree * bsizeKB;
    fs->val[VA_USED]  = (buf.f_blocks - buf.f_bfree) * bsizeKB;

#ifdef DEBUG
	syslog (LOG_DEBUG,
            "fetch_pathspace(): path %s total=%d used=%d free=%d available=%d\n",
            fs->path,
            fs->val[VA_TOTAL],
            fs->val[VA_USED],
            fs->val[VA_FREE],
            fs->val[VA_AVAIL]
            );
#endif	
}

void fetch_fs (sf_database *db)
{
    sf_list         *hd = db->fs;
    sf_fs_stats     *fs;

    // Iterate through list of watched paths
    while (hd) {
        fs = (sf_fs_stats *) hd->el;
        if (fs)
            fetch_pathspace (fs);
        hd = hd->next;
    }
}

void fetch_proc (sf_database *db)
{
    return;
}


void open_files ()
{
    lafha = open (lafile, O_RDONLY);
    if (lafha < 0) bail_out (EXIT_IO, lafile);
    
    memfha = open (memfile, O_RDONLY);
    if (memfha < 0) bail_out (EXIT_IO, memfile);
}

/* fast text parsing helpers */

char * fto_newline (char *b)
{
    do {
        b++;
    } while ((*b != '\n') && 
             (*b != '\r') && 
             (*b != '\0'));
    /* skip to new line */
    if (*b != '\0') b++;
    return b;
}

char * fto_notspace (char *b)
{
    do {
        b++;
    } while ((*b == ' ') || 
             (*b == '\t') || 
             (*b == '\n') || 
             (*b == '\r'));
    return b;
}

char * fto_space (char *b)
{
    do {
        b++;
    } while ((*b != ' ') && 
             (*b != '\t') && 
             (*b != '\n') && 
             (*b != '\r') &&
             (*b != '\0'));
    return b;
}


/* $Id: getstats.c,v 1.14 2004/05/29 12:55:09 emes Exp $ */
